
__author__ = 'Lukas Leufen'
__date__ = '2020-07-08'


from mlair.helpers import to_list, remove_items
import numpy as np
import xarray as xr
import pickle
import os
import pandas as pd
import datetime as dt
import shutil
import inspect
import copy

from typing import Union, List, Tuple, Dict
import logging
from functools import reduce
from mlair.data_handler.station_preparation import StationPrep
from mlair.helpers.join import EmptyQueryResult


number = Union[float, int]
num_or_list = Union[number, List[number]]


class DummyDataSingleStation:  # pragma: no cover

    def __init__(self, name, number_of_samples=None):
        self.name = name
        self.number_of_samples = number_of_samples if number_of_samples is not None else np.random.randint(100, 150)

    def get_X(self):
        X1 = np.random.randint(0, 10, size=(self.number_of_samples, 14, 5))  # samples, window, variables
        datelist = pd.date_range(dt.datetime.today().date(), periods=self.number_of_samples, freq="H").tolist()
        return xr.DataArray(X1, dims=['datetime', 'window', 'variables'], coords={"datetime": datelist,
                                                                                  "window": range(14),
                                                                                  "variables": range(5)})

    def get_Y(self):
        Y1 = np.round(0.5 * np.random.randn(self.number_of_samples, 5, 1), 1)  # samples, window, variables
        datelist = pd.date_range(dt.datetime.today().date(), periods=self.number_of_samples, freq="H").tolist()
        return xr.DataArray(Y1, dims=['datetime', 'window', 'variables'], coords={"datetime": datelist,
                                                                                  "window": range(5),
                                                                                  "variables": range(1)})

    def __str__(self):
        return self.name


class AbstractDataPreparation:

    _requirements = []

    def __init__(self, *args, **kwargs):
        pass

    @classmethod
    def build(cls, *args, **kwargs):
        """Return initialised class."""
        return cls(*args, **kwargs)

    @classmethod
    def requirements(cls):
        """Return requirements and own arguments without duplicates."""
        return list(set(cls._requirements + cls.own_args()))

    @classmethod
    def own_args(cls, *args):
        return remove_items(inspect.getfullargspec(cls).args, ["self"] + list(args))

    @classmethod
    def transformation(cls, *args, **kwargs):
        return None

    def get_X(self, upsampling=False, as_numpy=False):
        raise NotImplementedError

    def get_Y(self, upsampling=False, as_numpy=False):
        raise NotImplementedError

    def get_data(self, upsampling=False, as_numpy=False):
        return self.get_X(upsampling, as_numpy), self.get_Y(upsampling, as_numpy)

    def get_coordinates(self) -> Union[None, Dict]:
        return None


class DefaultDataPreparation(AbstractDataPreparation):

    _requirements = remove_items(inspect.getfullargspec(StationPrep).args, ["self", "station"])

    def __init__(self, id_class, data_path, min_length=0,
                 extreme_values: num_or_list = None, extremes_on_right_tail_only: bool = False, name_affix=None):
        super().__init__()
        self.id_class = id_class
        self.interpolation_dim = "datetime"
        self.min_length = min_length
        self._X = None
        self._Y = None
        self._X_extreme = None
        self._Y_extreme = None
        _name_affix = str(f"{str(self.id_class)}_{name_affix}" if name_affix is not None else id(self))
        self._save_file = os.path.join(data_path, f"data_preparation_{_name_affix}.pickle")
        self._collection = self._create_collection()
        self.harmonise_X()
        self.multiply_extremes(extreme_values, extremes_on_right_tail_only, dim=self.interpolation_dim)
        self._store(fresh_store=True)

    @classmethod
    def build(cls, station, **kwargs):
        sp_keys = {k: copy.deepcopy(kwargs[k]) for k in cls._requirements if k in kwargs}
        sp = StationPrep(station, **sp_keys)
        dp_args = {k: copy.deepcopy(kwargs[k]) for k in cls.own_args("id_class") if k in kwargs}
        return cls(sp, **dp_args)

    def _create_collection(self):
        return [self.id_class]

    @classmethod
    def requirements(cls):
        return remove_items(super().requirements(), "id_class")

    def _reset_data(self):
        self._X, self._Y, self._X_extreme, self._Y_extreme = None, None, None, None

    def _cleanup(self):
        directory = os.path.dirname(self._save_file)
        if os.path.exists(directory) is False:
            os.makedirs(directory)
        if os.path.exists(self._save_file):
            shutil.rmtree(self._save_file, ignore_errors=True)

    def _store(self, fresh_store=False):
        self._cleanup() if fresh_store is True else None
        data = {"X": self._X, "Y": self._Y, "X_extreme": self._X_extreme, "Y_extreme": self._Y_extreme}
        with open(self._save_file, "wb") as f:
            pickle.dump(data, f)
        logging.debug(f"save pickle data to {self._save_file}")
        self._reset_data()

    def _load(self):
        try:
            with open(self._save_file, "rb") as f:
                data = pickle.load(f)
            logging.debug(f"load pickle data from {self._save_file}")
            self._X, self._Y = data["X"], data["Y"]
            self._X_extreme, self._Y_extreme = data["X_extreme"], data["Y_extreme"]
        except FileNotFoundError:
            pass

    def get_data(self, upsampling=False, as_numpy=True):
        self._load()
        X = self.get_X(upsampling, as_numpy)
        Y = self.get_Y(upsampling, as_numpy)
        self._reset_data()
        return X, Y

    def __repr__(self):
        return ";".join(list(map(lambda x: str(x), self._collection)))

    def get_X_original(self):
        X = []
        for data in self._collection:
            X.append(data.get_X())
        return X

    def get_Y_original(self):
        Y = self._collection[0].get_Y()
        return Y

    @staticmethod
    def _to_numpy(d):
        return list(map(lambda x: np.copy(x), d))

    def get_X(self, upsampling=False, as_numpy=True):
        no_data = (self._X is None)
        self._load() if no_data is True else None
        X = self._X if upsampling is False else self._X_extreme
        self._reset_data() if no_data is True else None
        return self._to_numpy(X) if as_numpy is True else X

    def get_Y(self, upsampling=False, as_numpy=True):
        no_data = (self._Y is None)
        self._load() if no_data is True else None
        Y = self._Y if upsampling is False else self._Y_extreme
        self._reset_data() if no_data is True else None
        return self._to_numpy([Y]) if as_numpy is True else Y

    def harmonise_X(self):
        X_original, Y_original = self.get_X_original(), self.get_Y_original()
        dim = self.interpolation_dim
        intersect = reduce(np.intersect1d, map(lambda x: x.coords[dim].values, X_original))
        if len(intersect) < max(self.min_length, 1):
            X, Y = None, None
        else:
            X = list(map(lambda x: x.sel({dim: intersect}), X_original))
            Y = Y_original.sel({dim: intersect})
        self._X, self._Y = X, Y

    def get_observation(self):
        return self.id_class.observation.copy().squeeze()

    def get_transformation_Y(self):
        return self.id_class.get_transformation_information()

    def multiply_extremes(self, extreme_values: num_or_list = 1., extremes_on_right_tail_only: bool = False,
                          timedelta: Tuple[int, str] = (1, 'm'), dim="datetime"):
        """
        Multiply extremes.

        This method extracts extreme values from self.labels which are defined in the argument extreme_values. One can
        also decide only to extract extremes on the right tail of the distribution. When extreme_values is a list of
        floats/ints all values larger (and smaller than negative extreme_values; extraction is performed in standardised
        space) than are extracted iteratively. If for example extreme_values = [1.,2.] then a value of 1.5 would be
        extracted once (for 0th entry in list), while a 2.5 would be extracted twice (once for each entry). Timedelta is
        used to mark those extracted values by adding one min to each timestamp. As TOAR Data are hourly one can
        identify those "artificial" data points later easily. Extreme inputs and labels are stored in
        self.extremes_history and self.extreme_labels, respectively.

        :param extreme_values: user definition of extreme
        :param extremes_on_right_tail_only: if False also multiply values which are smaller then -extreme_values,
            if True only extract values larger than extreme_values
        :param timedelta: used as arguments for np.timedelta in order to mark extreme values on datetime
        """
        # check if X or Y is None
        if (self._X is None) or (self._Y is None):
            logging.debug(f"{str(self.id_class)} has no data for X or Y, skip multiply extremes")
            return
        if extreme_values is None:
            logging.debug(f"No extreme values given, skip multiply extremes")
            self._X_extreme, self._Y_extreme = self._X, self._Y
            return

        # check type if inputs
        extreme_values = to_list(extreme_values)
        for i in extreme_values:
            if not isinstance(i, number.__args__):
                raise TypeError(f"Elements of list extreme_values have to be {number.__args__}, but at least element "
                                f"{i} is type {type(i)}")

        for extr_val in sorted(extreme_values):
            # check if some extreme values are already extracted
            if (self._X_extreme is None) or (self._Y_extreme is None):
                X = self._X
                Y = self._Y
            else:  # one extr value iteration is done already: self.extremes_label is NOT None...
                X = self._X_extreme
                Y = self._Y_extreme

            # extract extremes based on occurrence in labels
            other_dims = remove_items(list(Y.dims), dim)
            if extremes_on_right_tail_only:
                extreme_idx = (Y > extr_val).any(dim=other_dims)
            else:
                extreme_idx = xr.concat([(Y < -extr_val).any(dim=other_dims[0]),
                                           (Y > extr_val).any(dim=other_dims[0])],
                                          dim=other_dims[1]).any(dim=other_dims[1])

            extremes_X = list(map(lambda x: x.sel(**{dim: extreme_idx}), X))
            self._add_timedelta(extremes_X, dim, timedelta)
            # extremes_X = list(map(lambda x: x.coords[dim].values + np.timedelta64(*timedelta), extremes_X))

            extremes_Y = Y.sel(**{dim: extreme_idx})
            extremes_Y.coords[dim].values += np.timedelta64(*timedelta)

            self._Y_extreme = xr.concat([Y, extremes_Y], dim=dim)
            self._X_extreme = list(map(lambda x1, x2: xr.concat([x1, x2], dim=dim), X, extremes_X))

    @staticmethod
    def _add_timedelta(data, dim, timedelta):
        for d in data:
            d.coords[dim].values += np.timedelta64(*timedelta)

    @classmethod
    def transformation(cls, set_stations, **kwargs):
        sp_keys = {k: copy.deepcopy(kwargs[k]) for k in cls._requirements if k in kwargs}
        transformation_dict = sp_keys.pop("transformation")
        if transformation_dict is None:
            return
        scope = transformation_dict.pop("scope")
        method = transformation_dict.pop("method")
        if transformation_dict.pop("mean", None) is not None:
            return
        mean, std = None, None
        for station in set_stations:
            try:
                sp = StationPrep(station, transformation={"method": method}, **sp_keys)
                mean = sp.mean.copy(deep=True) if mean is None else mean.combine_first(sp.mean)
                std = sp.std.copy(deep=True) if std is None else std.combine_first(sp.std)
            except (AttributeError, EmptyQueryResult):
                continue
        if mean is None:
            return None
        mean_estimated = mean.mean("Stations")
        std_estimated = std.mean("Stations")
        return {"scope": scope, "method": method, "mean": mean_estimated, "std": std_estimated}

    def get_coordinates(self):
        return self.id_class.get_coordinates()


def run_data_prep():

    from .data_preparation_neighbors import DataPreparationNeighbors
    data = DummyDataSingleStation("main_class")
    data.get_X()
    data.get_Y()

    path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "testdata")
    data_prep = DataPreparationNeighbors(DummyDataSingleStation("main_class"),
                                         path,
                                         neighbors=[DummyDataSingleStation("neighbor1"),
                                                    DummyDataSingleStation("neighbor2")],
                                         extreme_values=[1., 1.2])
    data_prep.get_data(upsampling=False)


def create_data_prep():

    from .data_preparation_neighbors import DataPreparationNeighbors
    path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "testdata")
    station_type = None
    network = 'UBA'
    sampling = 'daily'
    target_dim = 'variables'
    target_var = 'o3'
    interpolation_dim = 'datetime'
    window_history_size = 7
    window_lead_time = 3
    central_station = StationPrep("DEBW011", path, {'o3': 'dma8eu', 'temp': 'maximum'}, {},station_type, network, sampling, target_dim,
                                  target_var, interpolation_dim, window_history_size, window_lead_time)
    neighbor1 = StationPrep("DEBW013", path, {'o3': 'dma8eu', 'temp-rea-miub': 'maximum'}, {},station_type, network, sampling, target_dim,
                                  target_var, interpolation_dim, window_history_size, window_lead_time)
    neighbor2 = StationPrep("DEBW034", path, {'o3': 'dma8eu', 'temp': 'maximum'}, {}, station_type, network, sampling, target_dim,
                                  target_var, interpolation_dim, window_history_size, window_lead_time)

    data_prep = []
    data_prep.append(DataPreparationNeighbors(central_station, path, neighbors=[neighbor1, neighbor2]))
    data_prep.append(DataPreparationNeighbors(neighbor1, path, neighbors=[central_station, neighbor2]))
    data_prep.append(DataPreparationNeighbors(neighbor2, path, neighbors=[neighbor1, central_station]))
    return data_prep


if __name__ == "__main__":
    from mlair.data_handler.station_preparation import StationPrep
    from mlair.data_handler.iterator import KerasIterator, DataCollection
    data_prep = create_data_prep()
    data_collection = DataCollection(data_prep)
    for data in data_collection:
        print(data)
    path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "testdata", "keras")
    keras_it = KerasIterator(data_collection, 100, path, upsampling=True)
    keras_it[2]

