"""
Collections of bootstrap methods and classes.

How to use
----------

test

"""

__author__ = 'Felix Kleinert, Lukas Leufen'
__date__ = '2020-02-07'


import os
from collections import Iterator, Iterable
from itertools import chain

import numpy as np
import xarray as xr

from mlair.data_handler.advanced_data_handler import AbstractDataPreparation


class BootstrapIterator(Iterator):

    _position: int = None

    def __init__(self, data: "BootStraps"):
        assert isinstance(data, BootStraps)
        self._data = data
        self._dimension = data.bootstrap_dimension
        self._collection = self._data.bootstraps()
        self._position = 0

    def __next__(self):
        """Return next element or stop iteration."""
        try:
            index, dimension = self._collection[self._position]
            nboot = self._data.number_of_bootstraps
            _X, _Y = self._data.data.get_data(as_numpy=False)
            _X = list(map(lambda x: x.expand_dims({'boots': range(nboot)}, axis=-1), _X))
            _Y = _Y.expand_dims({"boots": range(nboot)}, axis=-1)
            single_variable = _X[index].sel({self._dimension: [dimension]})
            shuffled_variable = self.shuffle(single_variable.values)
            shuffled_data = xr.DataArray(shuffled_variable, coords=single_variable.coords, dims=single_variable.dims)
            _X[index] = shuffled_data.combine_first(_X[index]).reindex_like(_X[index])
            self._position += 1
        except IndexError:
            raise StopIteration()
        _X, _Y = self._to_numpy(_X), self._to_numpy(_Y)
        return self._reshape(_X), self._reshape(_Y), (index, dimension)

    @staticmethod
    def _reshape(d):
        if isinstance(d, list):
            return list(map(lambda x: np.rollaxis(x, -1, 0).reshape(x.shape[0] * x.shape[-1], *x.shape[1:-1]), d))
        else:
            shape = d.shape
            return np.rollaxis(d, -1, 0).reshape(shape[0] * shape[-1], *shape[1:-1])

    @staticmethod
    def _to_numpy(d):
        if isinstance(d, list):
            return list(map(lambda x: x.values, d))
        else:
            return d.values

    @staticmethod
    def shuffle(data: np.ndarray) -> np.ndarray:
        """
        Shuffle randomly from given data (draw elements with replacement).

        :param data: data to shuffle
        :return: shuffled data as numpy array
        """
        size = data.shape
        return np.random.choice(data.reshape(-1, ), size=size)


class BootStraps(Iterable):
    """
    Main class to perform bootstrap operations.

    This class requires a data handler following the definition of the AbstractDataPreparation, the number of bootstraps
    to create and the dimension along this bootstrapping is performed (default dimension is `variables`).

    When iterating on this class, it returns the bootstrapped X, Y and a tuple with (position of variable in X, name of
    this variable). The tuple is interesting if X consists on mutliple input streams X_i (e.g. two or more stations)
    because it shows which variable of which input X_i has been bootstrapped. All bootstrap combinations can be
    retrieved by calling the .bootstraps() method. Further more, by calling the .get_orig_prediction() this class
    imitates according to the set number of bootstraps the original prediction
    """
    def __init__(self, data: AbstractDataPreparation, number_of_bootstraps: int = 10,
                 bootstrap_dimension: str = "variables"):
        """
        Create iterable class to be ready to iter.

        :param data: a data generator object to get data / history
        :param number_of_bootstraps: the number of bootstrap realisations
        """
        self.data = data
        self.number_of_bootstraps = number_of_bootstraps
        self.bootstrap_dimension = bootstrap_dimension

    def __iter__(self):
        return BootstrapIterator(self)

    def __len__(self):
        return len(self.bootstraps())

    def bootstraps(self):
        l = []
        for i, x in enumerate(self.data.get_X(as_numpy=False)):
            l.append(list(map(lambda y: (i, y), x.indexes['variables'])))
        return list(chain(*l))

    def get_orig_prediction(self, path: str, file_name: str, prediction_name: str = "CNN") -> np.ndarray:
        """
        Repeat predictions from given file(_name) in path by the number of boots.

        :param path: path to file
        :param file_name: file name
        :param prediction_name: name of the prediction to select from loaded file (default CNN)
        :return: repeated predictions
        """
        file = os.path.join(path, file_name)
        prediction = xr.open_dataarray(file).sel(type=prediction_name).squeeze()
        vals = np.tile(prediction.data, (self.number_of_bootstraps, 1))
        return vals[~np.isnan(vals).any(axis=1), :]
