"""Implementation of run environment."""

__author__ = "Lukas Leufen"
__date__ = '2019-11-25'

import json
import logging
import os
import shutil
import time

from mlair.helpers.datastore import DataStoreByScope as DataStoreObject
from mlair.helpers.datastore import NameNotFoundInDataStore
from mlair.helpers import Logger
from mlair.helpers import TimeTracking
from mlair.plotting.tracker_plot import TrackPlot


class RunEnvironment(object):
    """
    Basic run class to measure execution time.

    Either call this class by 'with' statement or delete the class instance after finishing the measurement. The
    duration result is logged.

    .. code-block:: python

        >>> with RunEnvironment():
                <your code>
        INFO: RunEnvironment started
        ...
        INFO: RunEnvironment finished after 00:00:04 (hh:mm:ss)

    If you want to embed your custom module in a RunEnvironment, you can easily call it inside the with statement. If
    you want to exchange between different modules in addition, create your module as inheritance of the RunEnvironment
    and call it after you initialised the RunEnvironment itself.

    .. code-block:: python

        class CustomClass(RunEnvironment):

            def __init__(self):
                super().__init__()
                ...
            ...


        >>> with RunEnvironment():
                CustomClass()
        INFO: RunEnvironment started
        INFO: CustomClass started
        INFO: CustomClass finished after 00:00:04 (hh:mm:ss)
        INFO: RunEnvironment finished after 00:00:04 (hh:mm:ss)

    All data that is stored in the data store will be available for all other modules that inherit from RunEnvironment
    as long the RunEnvironemnt base class is running. If the base class is deleted either by hand or on exit of the with
    statement, this storage is cleared.

    .. code-block:: python

        class CustomClassA(RunEnvironment):

            def __init__(self):
                super().__init__()
                self.data_store.set("testVar", 12)


        class CustomClassB(RunEnvironment):

            def __init__(self):
                super().__init__()
                self.test_var = self.data_store.get("testVar")
                logging.info(f"testVar = {self.test_var}")


        >>> with RunEnvironment():
                CustomClassA()
                CustomClassB()
        INFO: RunEnvironment started
        INFO: CustomClassA started
        INFO: CustomClassA finished after 00:00:01 (hh:mm:ss)
        INFO: CustomClassB started
        INFO: testVar = 12
        INFO: CustomClassB finished after 00:00:02 (hh:mm:ss)
        INFO: RunEnvironment finished after 00:00:03 (hh:mm:ss)

    """

    # set data store and logger (both are mutable!)
    del_by_exit = False
    data_store = None
    logger = None
    tracker_list = []

    def __init__(self):
        """Start time tracking automatically and logs as info."""
        if RunEnvironment.data_store is None:
            RunEnvironment.data_store = DataStoreObject()
        if RunEnvironment.logger is None:
            RunEnvironment.logger = Logger()
        self.time = TimeTracking()
        logging.info(f"{self.__class__.__name__} started")
        # atexit.register(self.__del__)
        self.data_store.tracker.append({})
        self.tracker_list.extend([{self.__class__.__name__: self.data_store.tracker[-1]}])

    def __del__(self):
        """
        Finalise class.

        Only stop time tracking, if not already called by exit method to prevent duplicated logging (__exit__ is always
        executed before __del__) it this class was used in a with statement. If instance is called as base class and
        not as inheritance from this class, log file is copied and data store is cleared.
        """
        if not self.del_by_exit:
            self.time.stop()
            logging.info(f"{self.__class__.__name__} finished after {self.time}")
            self.del_by_exit = True
            # copy log file and clear data store only if called as base class and not as super class
            if self.__class__.__name__ == "RunEnvironment":
                try:
                    self.__plot_tracking()
                    self.__save_tracking()
                    self.__move_log_file()
                except FileNotFoundError:
                    pass
                self.data_store.clear_data_store()

    def __enter__(self):
        """Enter run environment."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit run environment."""
        if exc_type:
            logging.error(exc_val, exc_info=(exc_type, exc_val, exc_tb))
        self.__del__()

    def __move_log_file(self):
        try:
            new_file = self.__find_file_pattern("logging_%03i.log")
            logging.info(f"Move log file to {new_file}")
            shutil.move(self.logger.log_file, new_file)
            try:
                os.rmdir(os.path.dirname(self.logger.log_file))
            except (OSError, FileNotFoundError):
                pass
        except (NameNotFoundInDataStore, FileNotFoundError):
            pass

    def __save_tracking(self):
        tracker = self.data_store.tracker
        new_file = self.__find_file_pattern("tracking_%03i.json")
        logging.info(f"Copy tracker file to {new_file}")
        with open(new_file, "w") as f:
            json.dump(tracker, f)

    def __plot_tracking(self):
        plot_folder, plot_name = os.path.split(self.__find_file_pattern("tracking_%03i.pdf"))
        TrackPlot(self.tracker_list, sparse_conn_mode=True, plot_folder=plot_folder, plot_name=plot_name)

    def __find_file_pattern(self, name):
        counter = 0
        filename_pattern = os.path.join(self.data_store.get_default("logging_path", os.path.realpath(".")), name)
        new_file = filename_pattern % counter
        while os.path.exists(new_file):
            counter += 1
            new_file = filename_pattern % counter
        return new_file

    @staticmethod
    def do_stuff(length=2):
        """Just a placeholder method for testing without any sense."""
        time.sleep(length)
