"""Abstract workflow."""

__author__ = "Lukas Leufen"
__date__ = '2020-06-26'

from collections import OrderedDict

from mlair import RunEnvironment


class Workflow:
    """Abstract workflow class to handle sequence of stages (run modules). An inheriting class has to first initialise
    this mother class and can afterwards add an arbitrary number of stages by using the add method. The execution order
    is equal to the ordering of the stages have been added. To run the workflow, finally, a single call of the run
    method is sufficient. It must be taken care for inter-stage dependencies, this workflow class only handles the
    execution but not the dependencies (workflow would probably fail in this case)."""

    def __init__(self):
        self._registry = OrderedDict()

    def add(self, stage, **kwargs):
        """Add a new stage with optional kwargs."""
        self._registry[stage] = kwargs

    def run(self):
        """Run workflow embedded in a run environment and according to the stage's ordering."""
        with RunEnvironment():
            for stage, kwargs in self._registry.items():
                stage(**kwargs)
